#ifndef __LARCV3PROCESSOR_PROCESSORTYPES_H__
#define __LARCV3PROCESSOR_PROCESSORTYPES_H__

namespace larcv3 {

  typedef size_t ProcessID_t; ///< A unique identifier for an instantiated processes

}
#endif
