//
// cint script to generate libraries
// Declaire namespace & classes you defined
// #pragma statement: order matters! Google it ;)
//

#ifdef __CINT__
#pragma link off all globals;
#pragma link off all classes;
#pragma link off all functions;

#pragma link C++ class larcv::HistADC+;
//#pragma link C++ class larcv::SegImageAna+;
#pragma link C++ class larcv::HistTensor+;
//ADD_NEW_CLASS ... do not change this line
#endif























