import unittest

# def test_logger():
#     from larcv import base
#     # logger = base.logger.get_shared()
#     for level in [base.kDEBUG, base.kINFO, base.kNORMAL, base.kWARNING, base.kERROR]:
#         base.logger.force_level(level)

#         assert(level == base.logger.get_shared().level())