#ifndef __LARCV3PROCESSOR_PROCESSFACTORY_CXX__
#define __LARCV3PROCESSOR_PROCESSFACTORY_CXX__

#include "larcv3/core/processor/ProcessFactory.h"

larcv3::ProcessFactory* larcv3::ProcessFactory::_me = nullptr;

#endif
