/**
 * \file load_pyutil.h
 *
 * \ingroup PyUtil
 * 
 * \brief Class def header for a class load_pyutil
 *
 * @author kazuhiro
 */

/** \addtogroup PyUtil

    @{*/
#ifndef __LARCV3PYUTIL_LOAD_PYUTIL_H
#define __LARCV3PYUTIL_LOAD_PYUTIL_H

#include <iostream>

namespace larcv3 {
  /**
     \class load_pyutil
     User defined class load_pyutil ... these comments are used to generate
     doxygen documentation!
  */
  class load_pyutil{
    
  public:
    
    /// Default constructor
    load_pyutil();

    /// Default destructor
    ~load_pyutil(){}
    
  };
}

#endif
/** @} */ // end of doxygen group 

