//
// cint script to generate libraries
// Declaire namespace & classes you defined
// #pragma statement: order matters! Google it ;)
//

#ifdef __CINT__
#pragma link off all globals;
#pragma link off all classes;
#pragma link off all functions;

#pragma link C++ class larcv::ImageHolder+;
//#pragma link C++ class larcv::MCNuStream+;
#pragma link C++ class larcv::DataStream+;
#pragma link C++ class larcv::ImageMerger+;
#pragma link C++ class larcv::MergeTwoStream+;
//ADD_NEW_CLASS ... do not change this line
#endif




















